# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

TASK_AUTO = 'auto'
TASK_BINARY = 'binary'
TASK_MULTICLASS = 'multiclass'
TASK_REGRESSION = 'regression'
TASK_MULTILABEL = 'multilabel'


COLUMNNAME_POSTFIX_DISCRETE = '_discrete'
COLUMNNAME_POSTFIX_CATEGORIZE = '_cat'

# DATATYPE_TENSOR_FLOAT = 'float32'
# DATATYPE_PREDICT_CLASS = 'int32'
DATATYPE_LABEL = 'int16'