# -*- coding:utf-8 -*-
"""

"""


class Stateful:
    def __init__(self):
        pass

    def load_state(self):
        pass

    def save_state(self):
        pass

    def get_state_path(self):
        pass
