# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from hypernets.dispatchers.predict.grpc.proto import predict_pb2 as hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2


class PredictServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.predict = channel.unary_unary(
        '/hypernets.dispatchers.predict.grpc.proto.PredictService/predict',
        request_serializer=hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2.PredictRequest.SerializeToString,
        response_deserializer=hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2.PredictResponse.FromString,
        )


class PredictServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def predict(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PredictServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'predict': grpc.unary_unary_rpc_method_handler(
          servicer.predict,
          request_deserializer=hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2.PredictRequest.FromString,
          response_serializer=hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2.PredictResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'hypernets.dispatchers.predict.grpc.proto.PredictService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))
