# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hypernets/dispatchers/predict/grpc/proto/predict.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hypernets/dispatchers/predict/grpc/proto/predict.proto',
  package='hypernets.dispatchers.predict.grpc.proto',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n6hypernets/dispatchers/predict/grpc/proto/predict.proto\x12(hypernets.dispatchers.predict.grpc.proto\"8\n\x0ePredictRequest\x12\x11\n\tdata_file\x18\x01 \x01(\t\x12\x13\n\x0bresult_file\x18\x02 \x01(\t\"X\n\x0fPredictResponse\x12\x11\n\tdata_file\x18\x01 \x01(\t\x12\x13\n\x0bresult_file\x18\x02 \x01(\t\x12\x0c\n\x04\x63ode\x18\x03 \x01(\x05\x12\x0f\n\x07message\x18\x04 \x01(\t2\x93\x01\n\x0ePredictService\x12\x80\x01\n\x07predict\x12\x38.hypernets.dispatchers.predict.grpc.proto.PredictRequest\x1a\x39.hypernets.dispatchers.predict.grpc.proto.PredictResponse\"\x00\x62\x06proto3')
)




_PREDICTREQUEST = _descriptor.Descriptor(
  name='PredictRequest',
  full_name='hypernets.dispatchers.predict.grpc.proto.PredictRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_file', full_name='hypernets.dispatchers.predict.grpc.proto.PredictRequest.data_file', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_file', full_name='hypernets.dispatchers.predict.grpc.proto.PredictRequest.result_file', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=100,
  serialized_end=156,
)


_PREDICTRESPONSE = _descriptor.Descriptor(
  name='PredictResponse',
  full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_file', full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse.data_file', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_file', full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse.result_file', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='code', full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse.code', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse.message', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=158,
  serialized_end=246,
)

DESCRIPTOR.message_types_by_name['PredictRequest'] = _PREDICTREQUEST
DESCRIPTOR.message_types_by_name['PredictResponse'] = _PREDICTRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PredictRequest = _reflection.GeneratedProtocolMessageType('PredictRequest', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTREQUEST,
  '__module__' : 'hypernets.dispatchers.predict.grpc.proto.predict_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.predict.grpc.proto.PredictRequest)
  })
_sym_db.RegisterMessage(PredictRequest)

PredictResponse = _reflection.GeneratedProtocolMessageType('PredictResponse', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTRESPONSE,
  '__module__' : 'hypernets.dispatchers.predict.grpc.proto.predict_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.predict.grpc.proto.PredictResponse)
  })
_sym_db.RegisterMessage(PredictResponse)



_PREDICTSERVICE = _descriptor.ServiceDescriptor(
  name='PredictService',
  full_name='hypernets.dispatchers.predict.grpc.proto.PredictService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=249,
  serialized_end=396,
  methods=[
  _descriptor.MethodDescriptor(
    name='predict',
    full_name='hypernets.dispatchers.predict.grpc.proto.PredictService.predict',
    index=0,
    containing_service=None,
    input_type=_PREDICTREQUEST,
    output_type=_PREDICTRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_PREDICTSERVICE)

DESCRIPTOR.services_by_name['PredictService'] = _PREDICTSERVICE

# @@protoc_insertion_point(module_scope)
